subroutine MOPsolver(solver,n,m,nC,x,l,u,equatn,linear,jcnnzmax,hnnzmax,scaleF,&
	              checkder,outiter,time,nfev,ngev,theta,inform)

	use globals , only: sF,nfinner,xinner,d,JF
	use myproblem

	implicit none
	
	! SCALARS ARGUMENTS
	integer,intent(in)       :: solver,n,m,nC,jcnnzmax,hnnzmax
	logical,intent(in)       :: scaleF,checkder
	integer,intent(out)      :: outiter,nfev,ngev,inform
	real(kind=4),intent(out) :: time
	real(kind=8),intent(out) :: theta
	
	! ARRAY ARGUMENTS
	real(kind=8),intent(inout) :: x(n)
	real(kind=8),intent(in)    :: l(n),u(n)
	logical,intent(in)         :: equatn(nC),linear(nC)
	
	! LOCAL SCALARS
	integer :: ind,infoLS,infoSD,infoFG,allocerr,maxoutiter,nfevLS				 
	real(kind=8) :: stp,stpmin,epsopt,ftol             
	real(kind=4) :: timeini,timefin
	logical :: iprint,iprintLS

	! LOCAL ARRAYS
	real(kind=8) :: p(n),phi0(m),gphi0(m),Fx(m),xprev(n)

	! EXTERNAL SUBROUTINES
	external :: evalphi
	
  ! Given a  continuous differentiable function F:R^n->R^m and a convex 
  ! compact set C, this subroutine solves the multiobjective optimization 
  ! problem
  !
  ! Minimize F(x) subject to x in C,
  !
  ! using the Conditional Gradient Method described in
  !
  ! P. B. Assunção Filho, O. P. Ferreira, and L. F. Prudente, 
  ! Conditional gradient method for multiobjective optimization, 
  ! technical report, 2020
  !
  ! or the projected gradient method. Choose the solver to be used by
  ! taking:
  !
  ! solver = 0: projected gradient method
  ! solver = 1: Conditional Gradient Method
  ! 
  ! The user is required to code the function values and their gradients
  ! in subroutine myproblem.
  !
  ! The condition gradient direction is calculated in innersolver.f90 
  ! routine. This routine uses the software Algencan contained in 
  ! libalgencan.a. A step size satifying the Armijo condition is computed
  ! at each iteration by means of the line search routine armijo.f90. 
  !
  ! inform:
  !
  !  0: Optimality satisfied
  !  1: Maximum number of iterations reached
  ! -1: An error occurred during the execution of NLCG algorithm
  !
  ! March of 2020.
  !
  
  	!---------------------------------------------------------------------   
	!     Initialization
	!---------------------------------------------------------------------
		
	! Check derivatives
	
	if ( checkder ) call checkdF(n,m,x,l,u)
	
	! Start timing
		
	call cpu_time(timeini)
	
	! Set default parameters 

	epsopt = 5.0d0 * sqrt( 2.0d0 ** (-52) )
	ftol   = 1.d-04
		
	maxoutiter = 1000
	stpmin     = 1.0d-15
	
	! Checking the chosen solver
	if ( solver /= 0 .and. solver /= 1 ) then
		write(*,900)
		inform = - 1
		return
	end if
		
	! Allocating the arrays
	
	allocate(d(n),xinner(n),sF(m),JF(m,n),stat=allocerr)
	if ( allocerr .ne. 0 ) then
		write(*,*) 'Allocation error in main program'
		stop
	end if
	
	! Saving the inner informations
	
	nfinner = n
	xinner  = x
	
	! Print problem information	
	
	iprint   = .true.	
	iprintLS = .false.
	
	if ( iprint ) then
		if ( solver == 0 ) write(*,1000)
		if ( solver == 1 ) write(*,1001)
		write(*,1010) n, m
	end if
	
	! Scale problem
	
	call scalefactor(n,m,x,scaleF)	

	! Counters
	
	outiter = 0	
	nfev    = 0
	ngev    = 0
	
	!---------------------------------------------------------------------  
	!     Main loop
	!---------------------------------------------------------------------     

	Main_loop: do
	
		! Compute the Jacobian JF
		
		do ind = 1,m
			call sevalg(n,x,JF(ind,:),ind,infoFG)
			ngev = ngev + 1
			
			if ( infoFG /= 0 ) then
				inform = -1
				deallocate(d,xinner,sF,JF)
				return
			end if
		end do
	
		! Compute the search direction
		
		if ( solver == 0 ) call evalSD(n,m,nC,p,l,u,equatn,linear,jcnnzmax,hnnzmax,infoSD)
		if ( solver == 1 ) call evalCondGD(n,m,nC,p,l,u,equatn,linear,jcnnzmax,hnnzmax,infoSD)
		
		d = p - x

		! Compute theta
		
		if ( solver == 0 ) theta = maxval( matmul( JF, d ) ) + 0.5d0 * norm2( d ) ** 2
		if ( solver == 1 ) theta = maxval( matmul( JF, d ) )
	
		! Print information

		if ( iprint ) then
			if ( outiter == 0 ) then
				write(*,1030) epsopt
				write(*,1040)
				write(*,1050) outiter,abs(theta),infoSD,nfev,ngev
			else
				if ( mod(outiter,10) == 0 ) write(*,1040)
				write(*,1060) outiter,abs(theta),infoLS,infoSD,nfev,ngev
			end if
		end if
	
		!-------------------------------------------------------------------
		!     Test stopping criteria
		!-------------------------------------------------------------------
		
		! Test optimality	
		
		if ( abs( theta ) <= epsopt  ) then
	
			inform = 0
			
			! Stop timing
			
			call cpu_time(timefin)
			
			time = timefin - timeini
  
			if ( iprint ) write(*,1070) nfev, ngev, time
			deallocate(d,xinner,sF,JF)
			return
			
		end if		
		
		! Test whether the number of iterations is exhausted
		
		if ( outiter >= maxoutiter ) then
				inform = 1
				
				! Stop timing
			
				call cpu_time(timefin)
			
				time = timefin - timeini
					
				if ( iprint ) write(*,1080) nfev, ngev, time

				deallocate(d,xinner,sF,JF)
				return
		end if

		!-------------------------------------------------------------------
		!     Iteration
		!-------------------------------------------------------------------
		
		! Increment outiter
				
		outiter = outiter + 1
						
		! Compute initial step
		
		stp = 1.0d0
				
		! Compute phi0 = [phi_1(0),...,phi_m(0)] 
		
		if ( outiter == 1 ) then 
			do ind = 1, m
				call sevalf(n,x,phi0(ind),ind,infoFG)
				nfev = nfev + 1   				
			end do
		else
			phi0 = Fx
		end if
		
		! Compute gphi0 = [gphi_1(0),...,gphi_m(0)]
		
		gphi0 = matmul( JF, d )			
		
		! Compute the stepsize satisfying the Wolfe conditions 
											  
		call armijo(evalphi,stp,stpmin,m,phi0,gphi0,Fx,ftol,iprintLS,  &
		            nfevLS,infoLS) 
		            					
		! Check for errors in lsvecopt
		
		if ( infoLS == - 1 ) then
			inform = - 1

			deallocate(d,xinner,sF,JF)
			return
		end if
				 
		nfev = nfev + nfevLS
		
		! Update x	
		
		xprev = x
		
		x = x + stp * d
		
		! Saving the inner informations
		
		xinner  = x
		
	!---------------------------------------------------------------------
	!     Iterate
	!---------------------------------------------------------------------    

	end do Main_loop

	!--------------------------------------------------------------------- 
	!     End of main loop
	!---------------------------------------------------------------------
	
	! Non-executable statements
	
	900  format(/,1X,'Flag of MOPsolver: Invalid value for the solver variable',/)
	1000 format(/,1X,'-------------------------------------------------------------------------------',/,1X,& 
					 '           Projected Gradient Method for Multiobjective Optimization           ',/,1X,& 
					 '-------------------------------------------------------------------------------')
	1001 format(/,1X,'-------------------------------------------------------------------------------',/,1X,& 
					 '          Conditional Gradient Method for Multiobjective Optimization          ',/,1X,& 
					 '-------------------------------------------------------------------------------')
					 								 
	1010 format(/,1X,'Number of variables:',1X,I6,/,1X,'Number of functions:',1X,I6)	
	
	1030 format(/,1X,'Optimality tolerance:',1X,1P,D7.1)
	
	1040 format(/,4X,'out',2X,'|theta|',3X,'LS',1X,'IS',1X,'#evalf',2X,'#evalg')
	
	1050 format(1X,I5,3X,1P,D8.2,3X,'-',2X,I1,1X,I6,2X,I6)
	
	1060 format(1X,I5,3X,1P,D8.2,3X,I1,2X,I1,1X,I6,2X,I6)		
	
	1070 format(/,1X,'Flag of MOPsolver: solution was found',/,/,1X, &
					 'Number of functions evaluations:               ',I6,/,1X, &
					 'Number of derivatives evaluations:             ',I6/,/,1X, &
					 'Total CPU time in seconds: ',0P,F8.2)						
	
	1080 format(/,1X,'Flag of MOPsolver: maximum of iterations reached',/,/,1X,&
					 'Number of functions evaluations:               ',I6,/,1X, &
					 'Number of derivatives evaluations:             ',I6/,/,1X, &
					 'Total CPU time in seconds: ',0P,F8.2)

end subroutine MOPsolver
	
	
!***********************************************************************
!***********************************************************************	

subroutine scalefactor(n,m,x,scaleF)	

	use globals, only: sF
	use myproblem
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in)  :: n,m
	logical, intent(in)  :: scaleF

	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n)
		
	! LOCAL SCALAR
	integer :: i,ind
	real(kind=8) :: eps
	
	! LOCAL ARRAY
	real(kind=8) :: g(n)
	
	eps = 1.0d-08

! Compute objective functions scaling factor

	if ( scaleF ) then
		do ind = 1,m
			call evalg(n,x,g,ind)

			sF(ind) = 1.0d0
			do i = 1,n
				sF(ind) = max( sF(ind), abs( g(i) ) )
			end do
			
			sF(ind) = max( eps, 1.0d0 / sF(ind) )
		end do
	else
		sF(1:m) = 1.0d0
	end if
	
end subroutine scalefactor

!***********************************************************************
!***********************************************************************

subroutine sevalf(n,x,f,ind,inform)

	use globals, only: sF
	use myproblem

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in)  :: n,ind
	integer, intent(out) :: inform
	real(kind=8), intent(in) :: x(n)	
	real(kind=8), intent(out) :: f

	! FUNCTIONS
	logical :: IsANumber

	inform = 0 

	call evalf(n,x,f,ind)

	if ( .not. IsANumber(f) ) then
		inform = -1
		write(*,1000) ind
	end if

	f = sF(ind) * f
                
 1000	format(/,1X,'WARNING: The objective function value ',   &
                'computed by the user-supplied',/,1X,'subroutine ', &
                'evalf may be +Inf, -Inf or NaN. Function number:',1X,I4)	
                
end subroutine sevalf

!***********************************************************************
!***********************************************************************

subroutine sevalg(n,x,g,ind,inform)

	use globals, only: sF
	use myproblem
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in)  :: n,ind
	integer, intent(out) :: inform
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n)
	real(kind=8), intent(out) :: g(n)
	
	! LOCAL SCALAR
	integer :: i
	
	! FUNCTIONS
  logical :: IsANumber
  
  inform = 0 
	
	call evalg(n,x,g,ind)
	
	do i = 1,n
		if ( .not. IsANumber(g(i)) ) then
			inform = -1
			write(*,1000) ind,i
		end if
	end do
	
	g = sF(ind) * g
	
 1000	format(/,1X,'WARNING: There is an element whose value may be +Inf, -Inf or NaN in the gradient',/,10X,&
					'of the objective computed by the user-supplied subroutine evalg.',/,10X,& 
					'Function number:',1X,I4,2X,'Coordenate:',1X,I5)	
end subroutine sevalg


!***********************************************************************
!***********************************************************************

subroutine evalphi(stp,phi,ind)

	use globals, only: n => nfinner, x => xinner, d

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in) :: ind
	real(kind=8), intent(in) :: stp
	real(kind=8), intent(out) :: phi
	
	! LOCAL SCALAR
	integer :: infoFG

	call sevalf(n,x + stp * d,phi,ind,infoFG)

end subroutine evalphi

!***********************************************************************
!***********************************************************************

logical function IsANumber(x)

  implicit none

  ! SCALAR ARGUMENTS
  real(kind=8), intent(in) :: x
  
  ! LOCAL SCALAR
  real(kind=8) :: bignum
  
  bignum = 1.0d+99

  IsANumber = .true.
  if ( .not. abs( x ) .le. bignum ) IsANumber = .false.

end function IsANumber
